import os
import time
from collections import OrderedDict
import subprocess
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing
import warnings

warnings.filterwarnings("ignore", category=FutureWarning)
import copy

ker_model_size ={
    'r9RNA': 5,
    'r9DNA': 6,
    'r10DNA': 9,
    'rna004RNA': 9
}

base_shift_dict ={
    'r9RNA+': -1,
    'r9RNA-': -3,
    'r9DNA+': -2,
    'r9DNA-': -3,
    'r10DNA+': -6,
    'r10DNA-': -2,
    'rna004RNA+': -3,
    'rna004RNA-': -5,
}

def caculate_base_shift_size(kmer_model,strand):
    def is_odd(number):
        if number % 2 == 0:
            return False
        else:
            return True
    if kmer_model<=1:
        return 0
    if is_odd(kmer_model):
        return int((kmer_model-1)/2)
    elif strand=='+':
        return int((kmer_model ) / 2)-1
    else:
        return int((kmer_model) / 2)

def run_cmd(cmd):
    try:
        # 执行命令并捕获输出
        output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT, universal_newlines=True)

        # 处理命令输出
        print(output)
    except subprocess.CalledProcessError as e:
        # 处理命令执行错误
        print("Command execution failed:", e)
        raise RuntimeError('There are some errors in the cmd as below, please check your env\n'+cmd)
def read_fasta_to_dic(filename):
    """
    function used to parser small fasta
    still effective for genome level file
    """
    fa_dic = OrderedDict()

    with open(filename, "r") as f:
        for n, line in enumerate(f.readlines()):
            if line.startswith(">"):
                if n > 0:
                    fa_dic[short_name] = "".join(seq_l)  # store previous one

                full_name = line.strip().replace(">", "")
                short_name = full_name.split(" ")[0]
                seq_l = []
            else:  # collect the seq lines
                if len(line) > 8:  # min for fasta file is usually larger than 8
                    seq_line1 = line.strip()
                    seq_l.append(seq_line1)

        fa_dic[short_name] = "".join(seq_l)  # store the last one
    return fa_dic


def identify_file_path(file_path):
    if not os.path.exists(file_path):
        raise FileNotFoundError("File do not exist! Please check your path : " + file_path)


def generate_bam_file(fastq_file, reference, cpu,subsample_ratio=1):
    bam_file = '.'.join(fastq_file.split('.')[:-1]) + '_aligned.bam'
    if not os.path.exists(bam_file):
        cmds = 'minimap2 -ax map-ont -t ' + cpu + ' --MD ' + reference + ' ' + fastq_file + ' | samtools view -hbS -F ' + str(
            3332) + '  - | samtools sort -@ ' + cpu + ' -o ' + bam_file
        print('Start to alignment ...')
        run_cmd(cmds)
        print('bam file is saved in ' + bam_file)
    else:
        print(bam_file + ' existed. Will skip the minimap2 ... ')

    if subsample_ratio < 1:
        new_bam = '.'.join(fastq_file.split('.')[:-1]) + '_sub.bam'
        cmds = "samtools view -hbS -s " +str(subsample_ratio) +' ' + bam_file +' > ' + new_bam
        print(cmds)
        run_cmd(cmds)
        bam_file = new_bam
        cmds = 'samtools index ' + bam_file
        run_cmd(cmds)

    if not os.path.exists(bam_file+'.bai'):
        cmds = 'samtools index ' + bam_file
        run_cmd(cmds)

    new_fastq_file = '.'.join(bam_file.split('.')[:-1]) + '.fastq'
    if not os.path.exists(new_fastq_file):
        cmds = 'samtools bam2fq ' + bam_file + ' > '+ new_fastq_file
        run_cmd(cmds)
    return new_fastq_file,bam_file

def generate_paf_file_eventalign(fastq_file, blow5_file,bam_file,fasta_file,pore,rna,cpu):
    paf_file =  '.'.join(fastq_file.split('.')[:-1]) + '_ev.paf'
    if not os.path.exists(paf_file):
        if not os.path.exists(blow5_file+'.idx'):
            cmds = 'slow5tools index ' + blow5_file
            run_cmd(cmds)

        cmds = 'f5c index --slow5 ' +blow5_file+' '+ fastq_file
        run_cmd(cmds)

        cmds = 'f5c eventalign -r '+ fastq_file +" -g "+fasta_file+ ' --slow5 ' + blow5_file + ' --pore '+ pore+' -b ' + bam_file +' -c --min-mapq 0' + ' -t ' + str(cpu)
        if rna:
            cmds =cmds +' --rna'
        cmds =cmds +' > '+ paf_file
        print(cmds)
        print('Start to eventalign ...')
        run_cmd(cmds)
        print('Generated paf file : ' + paf_file)
    else:
        print(paf_file + ' existed. Will skip the f5c eventalign ... ')
    return paf_file

def generate_paf_file_resquiggle(fastq_file, blow5_file,pore,rna,cpu):
    paf_file =  '.'.join(fastq_file.split('.')[:-1]) + '_re.paf'
    if not os.path.exists(paf_file):
        if not os.path.exists(blow5_file + '.idx'):
            cmds = 'slow5tools index ' + blow5_file
            run_cmd(cmds)

        cmds = 'f5c index --slow5 ' +blow5_file+' '+ fastq_file
        run_cmd(cmds)

        cmds = 'f5c resquiggle -c ' + fastq_file + ' ' + blow5_file + ' --pore ' + pore + ' -o ' + paf_file +' -t '+ str(cpu)
        if rna:
            cmds =cmds +' --rna'
        print('Start to f5c resquiggle ...')
        run_cmd(cmds)
        print('Generated paf file : ' + paf_file)
    else:
        print(paf_file + ' existed. Will skip the f5c resquiggle ... ')
    return paf_file

def prepare_move_table_file(bam_file, reference, cpu,sig_move_offset,kmer_length):
    paf_file = '.'.join(bam_file.split('.')[:-1]) + '.paf'
    fastq_file = '.'.join(bam_file.split('.')[:-1]) + '.fastq'
    cmds = 'samtools index ' + bam_file
    run_cmd(cmds)
    print('Start to generate paf file  ...')
    cmds = 'squigualiser reform --sig_move_offset '+sig_move_offset+' --kmer_length '+kmer_length+' -c --bam ' + bam_file +' -o ' + paf_file
    print(cmds)
    run_cmd(cmds)
    print("Start alignment ...")
    cmds = 'samtools bam2fq '+bam_file+' >' + fastq_file
    run_cmd(cmds)
    aligned_fastq,aligned_bam = generate_bam_file(fastq_file, reference, cpu)
    return aligned_bam,paf_file


def run_samtools(fastq_file, location, reference, result_path, group, cpu):
    _,bam_file = generate_bam_file(fastq_file, reference, cpu)
    cmds = 'samtools mpileup ' + bam_file + ' -r ' + location + ' --no-output-ins --no-output-del -B -Q 0 -f ' + reference + ' -o ' + result_path + 'temp.txt'
    run_cmd(cmds)
    temp_file = pd.read_csv(result_path + 'temp.txt', sep='\t', header=None)
    temp_file['Group'] = group
    run_cmd('rm ' + result_path + 'temp.txt')
    return temp_file


def build_out_path(results_path):
    if not os.path.exists(results_path):
        os.mkdir(results_path)
    else:
        print("Output file existed! It will be overwrite or add content after 5 secs ...")
        time.sleep(5)
        print("Continue ...")

def reverse_fasta(ref):
    base_dict = {'A': 'T', "T": "A", "C": "G", "G": "C"}
    ref = np.array(list(ref.upper()))
    for index, element in enumerate(ref):
        ref[index] = base_dict[element]
    reverse_fasta = list(reversed(''.join(ref)))
    return ''.join(reverse_fasta)

def extract_kmer_feature(df_input, kmer, position):
    df = copy.deepcopy(df_input)
    def is_odd(number):
        if number % 2 == 1:
            return True
        else:
            return False
    if not is_odd(kmer) or kmer <= 0:
        raise Exception("The kmer should be an odd number and greater than zero.")

    kmer_size = (kmer-1)//2
    # df.loc[:, 'Dwell time'] = np.log10(df['Dwell time'])
    # df.loc[:, 'Dwell time'] = stats.zscore(df['Dwell time'])
    df = df[(df['Position'] >= position-kmer_size) & (df['Position'] <= position+kmer_size)]
    # df = df.sort_values(['Read ID', 'Position'], ascending=True)
    # df = df.reset_index(drop=True)
    df.loc[:, 'Dwell time'] = np.log10(df['Dwell time'])
    #df['Dwell time'] = np.log10(df['Dwell time'])
    # df.loc[:, 'Dwell time'] = stats.zscore(df['Dwell time'])
    # df.loc[:, 'STD'] = stats.zscore(df['STD'])
    grouped_df = df.groupby('Read ID')
    # df = df.apply(stats.zscore, axis=0)

    result_list=[]
    label_list=[]
    for key,temp in grouped_df:
        item = temp[['Mean','STD','Median','Dwell time']].values
        item = item.reshape(-1,).tolist()
        if len(item) < kmer * 4:
            continue
        if len(item) > kmer * 4:
            print(1)
        result_list.append(item)
        label_list.append([temp['Group'].values[0]])
    feature_matrix = pd.DataFrame(result_list)
    label = pd.DataFrame(label_list)
    return feature_matrix, label


def process_group(read_df,shift_size):
    # fill nan row
    read_df.reset_index(drop=True, inplace=True)
    positions = read_df['Position'].astype(int)
    missing_positions = read_df[positions.diff() > 1].index

    # 构建空行 DataFrame
    empty_df = pd.DataFrame(index=missing_positions, columns=read_df.columns)

    # 合并空行和原始数据
    merged_df = pd.concat([read_df, empty_df]).sort_index()

    data_df = merged_df[['Mean', 'STD', 'Median', 'Dwell time']]
    previous_row = data_df.shift(shift_size)
    next_row = data_df.shift(-shift_size)

    # 将当前行、前一行和后一行的数据合并成新的一行
    merged_data = pd.concat([previous_row, data_df, next_row], axis=1)
    merged_data.columns = range(merged_data.shape[1])
    merged_data['Position'] = read_df['Chrom'] + ":" + read_df['Position'].astype(str) + ':' + read_df['Strand']
    # 删除空的列并重塑成新的一行
    merged_data = merged_data.dropna(axis=0).reset_index(drop=True)
    return merged_data.values.tolist()

def extract_feature_parallel(df, kmer_size, label, cpu):
    def is_odd(number):
        if number % 2 == 1:
            return True
        else:
            return False

    if not is_odd(kmer_size) or kmer_size < 0:
        raise Exception("The kmer should be an odd number and greater than zero.")

    window_size = (kmer_size - 1) // 2
    df_group = df.groupby('Read ID')

    results_list=[]
    pool = multiprocessing.Pool(processes=cpu)
    pbar = tqdm(total=df_group.ngroups, position=0, leave=True,unit='reads')

    for key, read_df in df_group:
        result_per_read = pool.apply_async(process_group, args=(read_df,window_size,))
        results_list.append(result_per_read)
    pool.close()
    pool.join()

    final_df = []
    for item in results_list:
        item = item.get()
        if item is not None:
            final_df.extend(item)
        pbar.update(1)

    pbar.close()
    print("Merging the result table ...")
    final_df = pd.DataFrame(final_df)
    columns = list(range(final_df.shape[1]-1))
    columns.append('Position')
    final_df.columns = columns
    final_df['Label'] = label
    return final_df

def extract_feature(df, kmer_size, label):
    def is_odd(number):
        if number % 2 == 1:
            return True
        else:
            return False

    if not is_odd(kmer_size) or kmer_size < 0:
        raise Exception("The kmer should be an odd number and greater than zero.")

    shift_size = (kmer_size - 1) // 2
    df_group = df.groupby('Read ID')

    pbar = tqdm(total= df_group.ngroups, position=0, leave=True)

    final_df = pd.DataFrame()
    for key, read_df in df_group:
        # fill nan row
        read_df.reset_index(drop=True, inplace=True)
        positions = read_df['Position'].astype(int)
        missing_positions = read_df[positions.diff() > 1].index

        # 构建空行 DataFrame
        empty_df = pd.DataFrame(index=missing_positions, columns=read_df.columns)

        # 合并空行和原始数据
        merged_df = pd.concat([read_df, empty_df]).sort_index()

        data_df = merged_df[['Mean', 'STD', 'Median', 'Dwell time']]
        previous_row = data_df.shift(shift_size)
        next_row = data_df.shift(-shift_size)

        # 将当前行、前一行和后一行的数据合并成新的一行
        merged_data = pd.concat([previous_row, data_df, next_row], axis=1)
        merged_data.columns = range(merged_data.shape[1])
        merged_data['Position'] = read_df['Chrom'] + ":" + read_df['Position'] + ':' + read_df['Strand']
        # 删除空的列并重塑成新的一行
        merged_data = merged_data.dropna(axis=0).reset_index(drop=True)
        final_df = pd.concat([final_df, merged_data], axis=0)
        pbar.update(1)
    pbar.close()
    final_df['Label'] = label
    return final_df

def save_fasta_dict(fasta_dict, path):
    f = open(path, 'w+')
    for key, value in fasta_dict.items():
        f.write('>' + key + '\n')
        line = len(value) // 80 + 1
        for i in range(0, line):
            f.write(value[i * 80:(i + 1) * 80] + '\n')
        f.close()

def calculate_MANOVA_result(position,df,length_size,subsample_num=500,windows_len=10,kmer=3):
    print("Start to run the MANOVA analysis on target region ...")
    from statsmodels.multivariate.manova import MANOVA
    from sklearn.decomposition import PCA
    # import umap
    kmer_size =(kmer-1)//2
    methylation_list=list(range( position - length_size + kmer_size ,position + length_size +1-kmer_size))
    result_list=[]
    for item in methylation_list:
        # subsample the reads
        control = df[df['Group']=='Control']
        # if control.shape[0] > subsample_num*(2*windows_len+1):
        #     control=control.iloc[0:subsample_num*(2*windows_len+1), :]
        sample = df[df['Group'] == 'Sample']
        # if sample.shape[0] > control.shape[0] * 2:
        #     sample = sample.iloc[0:control.shape[0],:]
        df = pd.concat([control,sample],axis=0).reset_index(drop=True)
        feature,label = extract_kmer_feature(df,kmer,item)
        pca = PCA(n_components=2,whiten=True)
        try:
            new_df = pd.DataFrame(pca.fit_transform(feature))
        except Exception as e:
            print(1)
        # reducer = umap.UMAP(n_components=2)  # Create a UMAP object with 2 dimensions
        # new_df = reducer.fit_transform(feature)
        new_df = pd.concat([pd.DataFrame(new_df),label], axis =1)
        new_df.columns=['PC1','PC2','Group']
        if np.sum(new_df['Group']=='Sample') > 10 and np.sum(new_df['Group']=='Control') > 10:
            manova = MANOVA.from_formula('PC1 + PC2 ~ Group', data=new_df)
            # 执行多元方差分析
            results = manova.mv_test()
            pvalue = results.summary().tables[3].iloc[0,5]
            kmer_center = int((kmer -1)/2)
            mean_differ = feature[label[0]=='Sample'][kmer_center*4].median() - feature[label[0]=='Control'][kmer_center*4].median()
            result_list.append([item,pvalue,mean_differ])
        else:
            result_list.append([item, None])
    new_df = pd.DataFrame(result_list)
    new_df[1] = np.log10(new_df[1]) * (-1)
    new_df.columns = ['Position', 'P value(-log10)','Norm_differ']
    return new_df

# fasta=read_fasta_to_dic("../example/data/23S_rRNA.fasta")
# for key,value in fasta.items():
#     fasta[key]=reverse_fasta(value)
# save_fasta_dict(fasta,'../example/data/23S_rRNA_re.fasta')




import pandas as pd
import plotnine as p9
import numpy as np
from matplotlib import pyplot as plt

plt.rcParams['pdf.fonttype'] = 42
# plt.rcParams['font.sans-serif'] = ['Arial']
SIG_PCTL_RANGE = [0.01,0.99]


# def draw_boxplot(df,results_path,pos,base_list,title):
#     item_list = ['Mean', 'STD', 'Median', 'Dwell_time']
#     plot_list=[]
#     for item in item_list:
#
#         sig_min, sig_max = np.percentile(df[item], SIG_PCTL_RANGE)
#         sig_diff = sig_max - sig_min
#         ylim_tuple = (sig_min - sig_diff * 0.1, sig_max + sig_diff * 0.1)
#
#         plot = p9.ggplot(df, p9.aes(x='position', y=item, fill='type')) \
#                + p9.theme_bw() \
#                +p9.geom_boxplot( outlier_shape='',position=p9.position_dodge(0.9),size=0.25) \
#                + p9.scale_fill_manual(values={"Sample": "#ff6f91", "Control": "#7389af"}) \
#                + p9.scale_x_discrete(labels=list(base_list)) \
#                + p9.theme(
#                     figure_size=(6, 3),
#                     panel_grid_minor=p9.element_blank(),
#                     axis_text=p9.element_text(size=13),
#                     axis_title=p9.element_text(size=13),
#                     title=p9.element_text(size=13),
#                     legend_position='none')\
#                + p9.ylim(ylim_tuple)\
#                + p9.labs(title=title, x=str(pos + 1), y=item)
#
#         plot.save(filename=results_path + "/" + item + "_boxplot.pdf", dpi=300)
#
# def draw_violin(df,results_path,pos,base_list,title):
#
#     item_list = ['Mean', 'STD', 'Median', 'Dwell_time']
#     for item in item_list:
#
#         sig_min, sig_max = np.percentile(df[item], SIG_PCTL_RANGE)
#         sig_diff = sig_max - sig_min
#         ylim_tuple = (sig_min - sig_diff * 0.1, sig_max + sig_diff * 0.1)
#
#         plot = p9.ggplot(df, p9.aes(x='position', y=item, fill='type')) \
#                + p9.geom_violin(style='left-right',position=p9.position_dodge(0),color='none',width=1.5) \
#                + p9.theme_bw() \
#                + p9.scale_fill_manual(values={"Sample": "#ff6f91", "Control": "#7389af"}) \
#                + p9.scale_x_discrete(labels=list(base_list)) \
#                + p9.theme(
#             figure_size=(6, 3),
#             panel_grid_minor=p9.element_blank(),
#             axis_text=p9.element_text(size=13),
#             axis_title=p9.element_text(size=13),
#             title=p9.element_text(size=13),
#             legend_position='none'
#         )\
#         + p9.ylim(ylim_tuple)
#
#         plot = plot + p9.labs(title=title, x=str(pos + 1), y=item)
#         # plot.render_matplotlib()
#         plot.save(filename=results_path + "/" + item + "_violin.pdf", dpi=300)
# print(plot)

def current_plot(df, results_path, pos, base_list, title,filter):
    print("Start to plot current feature ...")
    item_list = ['Mean', 'STD', 'Median', 'Dwell time']
    len_plot = (len(base_list)-1)//2
    pos_list = list(range(-len_plot+pos,len_plot+1+pos))
    pos_list = [str(num) for num in pos_list]
    new_df = pd.DataFrame()
    for item in item_list:
        # collect data
        temp = df[[item, 'Position', 'Group']].copy()
        temp.columns = ['value', 'Position', 'Group']
        temp.loc[:, 'stats'] = item
        if filter:
            sig_min = temp['value'].quantile(SIG_PCTL_RANGE[0])
            sig_max = temp['value'].quantile(SIG_PCTL_RANGE[1])
            ylim_tuple = [sig_min , sig_max]
            temp = temp[(temp['value'] >= ylim_tuple[0]) & (temp['value'] <= ylim_tuple[1])]
        new_df = pd.concat([new_df, temp], axis=0)
    new_df['Position']=new_df['Position'].astype(str)
    plot = p9.ggplot(new_df, p9.aes(x='Position', y="value", fill='Group')) \
           + p9.theme_bw() \
           + p9.scale_fill_manual(values={"Sample": "#F57070", "Control": "#9F9F9F", "Single": "#a3abbd"}) \
           + p9.scale_x_discrete(breaks=pos_list, labels=base_list) \
           + p9.theme(
        figure_size=(8, 8),
        panel_grid_minor=p9.element_blank(),
        axis_text=p9.element_text(size=13),
        axis_title=p9.element_text(size=13),
        title=p9.element_text(size=13),
        legend_position='bottom',
        legend_title=p9.element_blank(),
        strip_text=p9.element_text(size=13),
        strip_background=p9.element_rect(alpha=0),

    ) \
           + p9.facet_grid('stats ~', scales='free_y')
    plot = plot + p9.labs(title=title, x=str(pos + 1), y='')

    if new_df['Group'].drop_duplicates().shape[0] == 1:
        print("control files do not exist, will turn to the single mode")
        plot2 = plot + p9.geom_violin(color='none', position=p9.position_dodge(0.9), width=1)
        plot2 = plot2 + p9.geom_boxplot(outlier_shape='', position=p9.position_dodge(0.9), size=0.2, width=0.1)
        plot2.save(filename=results_path + "/current_single.pdf", dpi=300)
    else:
        plot1 = plot + p9.geom_boxplot(outlier_shape='', position=p9.position_dodge(0.9), size=0.2, width=0.75,alpha = 0.8)
        plot1.save(filename=results_path + "/current_boxplot.pdf", dpi=300)
        plot2 = plot + p9.geom_violin(style='left-right', position=p9.position_dodge(0), color='none', width=1.5,alpha = 0.8)
        plot2.save(filename=results_path + "/current_violin.pdf", dpi=300)
    print('Figures are saved in '+results_path)


def draw_signal(df, start, base,start_index,end_index):
    df = pd.DataFrame(df)
    df.columns = ['raw']
    start=start[start_index:end_index]
    df=df[start[0]:start[-1]]
    df = df.reset_index()
    plot = p9.ggplot(df, p9.aes(x='index', y="raw")) \
           + p9.theme_bw() \
           + p9.geom_line() \
           + p9.theme(
        figure_size=(6, 3),
        panel_grid_minor=p9.element_blank(),
        axis_text=p9.element_text(size=13),
        axis_title=p9.element_text(size=13),
        title=p9.element_text(size=13),
        strip_text=p9.element_text(size=13),
        legend_position='none',
        strip_background=p9.element_rect(alpha=0)
    )
    # plot.save(filename="/home/zhguo/Dropbox/@labfiles/projects/GUO/ONT_showcase_tool/plot/signal.pdf", dpi=300)
    for item in start:
        plot = plot + p9.geom_vline(xintercept=item, linetype='dashed', color='red')
    print(plot)
    # plot.save(filename="/home/zhguo/Dropbox/@labfiles/projects/GUO/ONT_showcase_tool/plot/norm_signal_aligned.pdf", dpi=300)

def alignment_plot(final_feature,pos_list,base_list,title,pos,results_path):
    order_list = ['Match', 'A', 'T', 'C', 'G']
    category = pd.api.types.CategoricalDtype(categories=order_list, ordered=True)
    final_feature['variable'] = final_feature['variable'].astype(category)

    order_list = ['Sample', 'Control','Single']
    category = pd.api.types.CategoricalDtype(categories=order_list, ordered=True)
    final_feature['Group'] = final_feature['Group'].astype(category)

    plot = p9.ggplot(final_feature, p9.aes(x='Position', y="value", fill='variable')) \
           + p9.theme_bw() \
           + p9.geom_bar(stat='identity') \
           + p9.scale_x_continuous(breaks=pos_list, labels=base_list) \
           + p9.scale_fill_manual(values={"Match": "#D7D7D7",
                                          'T': "#C88776",
                                          'A': '#95B989',
                                          'G': '#FFCE9B',
                                          'C': '#7CB3C5'}) \
           + p9.theme(
        figure_size=(8, 4),
        panel_grid_minor=p9.element_blank(),
        axis_text=p9.element_text(size=13),
        axis_title=p9.element_text(size=13),
        title=p9.element_text(size=13),
        legend_position='bottom',
        legend_title=p9.element_blank(),
        strip_text=p9.element_text(size=13),
        strip_background=p9.element_rect(alpha=0),

    ) \
           + p9.facet_grid('Group ~', scales='free_y')
    plot = plot + p9.labs(title=title, x=str(pos), y='Coverage')
    plot.save(filename=results_path + "/alignment.pdf", dpi=300)
    print('Figures are saved in ' + results_path)

def plot_PCA(feature_matrix,label,results_path):
    print("Start to plot the PCA distribution on target position ...")

    from sklearn.decomposition import PCA
    pca = PCA(n_components=2)
    new_df = pd.DataFrame(pca.fit_transform(feature_matrix))
    weights = pca.explained_variance_ratio_

    # 打印权重
    print("Weights:", weights)

    x_axis = 'PC1 : ' + str(round(weights[0],2))
    y_axis = 'PC2 : ' + str(round(weights[1],2))
#     y_test = label[0].apply(lambda x: 1 if x == 'Sample' else 0)
#     # new_df.columns = ['PC1', 'PC2','Group']
#     # explained_variance_ratio = pca.explained_variance_ratio_
#     # for i, ratio in enumerate(explained_variance_ratio):
#     #     print(f"Principal Component {i + 1}: {ratio:.2f}")
#     new_df =  umap.UMAP(
#     n_neighbors=15,
#     min_dist=0.0,
#     n_components=2,
#     random_state=42,
# ).fit_transform(feature_matrix)
    # plt.scatter(new_df[:, 0], new_df[:, 1], c=y_test, s=0.1, cmap='Spectral');

    # reducer = umap.UMAP(n_components=2)  # 指定降维后的维度为2
    # new_df = reducer.fit_transform(feature_matrix)
    new_df = pd.concat([pd.DataFrame(new_df), label], axis=1)
    new_df.columns = ['PC1', 'PC2', 'Group']
    plot = p9.ggplot(new_df, p9.aes(x='PC1', y='PC2', color='Group')) \
           + p9.theme_bw() \
            +p9.labs(x=x_axis,y=y_axis)\
           + p9.scale_color_manual(values={"Sample": "#F57070", "Control": "#9F9F9F", "Single": "#a3abbd"}) \
           + p9.geom_point() \
           + p9.geom_density_2d() \
           + p9.theme(
        figure_size=(5, 5),
        panel_grid_minor=p9.element_blank(),
        axis_text=p9.element_text(size=13),
        axis_title=p9.element_text(size=13),
        title=p9.element_text(size=13),
        legend_position='bottom',
        legend_title=p9.element_blank(),
        strip_text=p9.element_text(size=13),
        strip_background=p9.element_rect(alpha=0),
    )
    # print(plot)
    plot.save(filename=results_path + "/PCA_target_position.pdf", dpi=300)

def MANOVA_plot(df, results_path,cutoff):

    df['Result'] = df['P value(-log10)'].apply(lambda x: 'Significant' if x >= cutoff else 'Not significant')
    df.to_csv(results_path + '/MANOVA_result.csv', index=None)
    print("Feature file saved  in " + results_path + '/MANOVA_result.csv')
    plot = p9.ggplot(df, p9.aes(x='Position', y='P value(-log10)', fill='Result')) \
           + p9.theme_bw() \
           + p9.geom_col() \
           + p9.scale_fill_manual(values={'Significant': "#BB9A8B", 'Not significant': "#D7D7DD"}) \
           + p9.geom_hline(yintercept=cutoff, linetype='dashed') \
           + p9.theme(
        figure_size=(8, 3),
        panel_grid_minor=p9.element_blank(),
        axis_text=p9.element_text(size=13),
        axis_title=p9.element_text(size=13),
        title=p9.element_text(size=13),
        legend_position='bottom',
        legend_title=p9.element_blank(),
        strip_text=p9.element_text(size=13),
        strip_background=p9.element_rect(alpha=0),
    )
    # plot.save(filename=results_path + "/zscore_density.pdf", dpi=300)
    plot.save(filename=results_path + "/MANOVA_distribution.pdf", dpi=300)


import re
import numpy as np
import pandas as pd
import pyslow5
import pysam
from tqdm import tqdm
from nanoCEM.normalization import normalize_signal,normalize_signal_with_lim
from nanoCEM.cem_utils import generate_bam_file,identify_file_path,prepare_move_table_file
# from nanoCEM.plot import draw_signal
import os
import argparse
score_dict={}
nucleotide_type=None
def extract_feature(line,strand,sig_move_offset,norm=True):
    pbar.update(1)
    read_id = line[0]
    # if read_id =='dd49b363-ab1e-45c0-ae06-7d7a06452f4c':
    #     print(1)
    if read_id not in info_dict:
        return None
    # tackle moves tag
    moves_string = line[12]
    moves_string = re.sub('ss:Z:', '', moves_string)
    moves_string = re.sub('D', 'D,', moves_string)
    moves_string = re.sub('I', 'I,', moves_string)
    # print(moves_string)
    moves = re.split(r',+', moves_string)
    moves = moves[:-1]
    # extract index and generate event_length and event_start
    insertion = 0
    event_length = []
    for i,item in enumerate(moves):
        if 'D' in item:
            deletion = int(item[:-1])
            for i in range(deletion):
                event_length.append(0)
        elif 'I' in item:
            if i == 0 :
                continue
            else:
                return None
        elif '=' in item:
            return None
        else:
            event_length.append(int(item))
    # build event_length from move table

    if read_id not in read_ids:
        return None
    read = s5.get_read(read_id, aux=["read_number", "start_mux"],pA=True)
    start_index = line[2]
    end_index = line[3]
    event_length = np.array(event_length)
    # assert len_raw_signal in paf and blow5

    try:
        assert read['len_raw_signal'] == line[1]
    except Exception:
        print("Warning: 1 read's length of signal is not equal between blow5 and paf")
        return None
    signal = read['signal']

    # create event start and flip all table
    signal = signal[start_index:end_index]

    event_starts = event_length.cumsum()
    event_starts = np.insert(event_starts, 0, 0)[:-1]
    if norm:
        signal,shift,scale = normalize_signal_with_lim(signal)

    # index query and reference
    aligned_pair=info_dict[read_id]['pairs']
    qlen = info_dict[read_id]['query_length']
    try:
        assert qlen == len(event_length) + sig_move_offset
    except Exception:
        print("Warning: 1 read's length of event is not equal between bam and paf file")
        return None
    # correct index about DNA and RNA
    if (nucleotide_type == 'RNA' and strand == '+') or (nucleotide_type == 'DNA' and strand == '-'):
        aligned_pair[0] = qlen - aligned_pair[0] - 1

    if aligned_pair.shape[0]==0:
        return None
    read_pos = aligned_pair[0].values
    ref_pos = aligned_pair[1].values

    # extract raw signal by event length and event start
    total_feature_per_reads = []
    raw_signal_every = [signal[event_starts[x]:event_starts[x] + event_length[x]] for x in
                        read_pos]
    # calculate mean median and dwell time
    for i, element in enumerate(raw_signal_every):
        if len(element) == 0:
            continue
        temp = [read_id,np.mean(element), np.std(element), np.median(element), event_length[read_pos[i]],ref_pos[i]]
        total_feature_per_reads.append(temp)
    return total_feature_per_reads

def extract_pairs_pos(bam_file,position,length,chromosome,strand):

    result_dict={}
    for read in bam_file.fetch(chromosome,position-length, position+length+1):
        if strand == '+' and read.is_reverse:
            continue
        if strand == '-' and not read.is_reverse:
            continue
        # if read.qname == '7dcec2bc-65f3-41fe-8981-dd3af9fa6e67':
        #     print(1)
        start_position=read.reference_start
        end_position=read.reference_end
        if position < start_position or position > end_position:
            continue
        # unit
        aligned_pair = pd.DataFrame(np.array(read.aligned_pairs)).dropna().reset_index(drop=True)
        aligned_pair = aligned_pair[(aligned_pair[1] >= position - length) & (aligned_pair[1] <= position + length)]

        temp={}
        temp['pairs']=aligned_pair
        temp['query_length'] = read.query_length
        result_dict[read.qname] = temp
    return result_dict



def read_basecall_bam(path,position,reference,length,chrom,strand,sig_move_offset,kmer_length,subsample_ratio=1,norm=True,cpu=4,rna=True):
    global info_dict, s5,pbar,nucleotide_type,read_ids
    slow5_file = path + ".blow5"
    bam_file = path + ".bam"
    identify_file_path(bam_file)
    identify_file_path(slow5_file)
    align_bam_file, paf_file = prepare_move_table_file(bam_file,reference,cpu,str(sig_move_offset),str(kmer_length))

    if rna:
        nucleotide_type ='RNA'
    else:
        nucleotide_type = 'DNA'
    bam_file = pysam.AlignmentFile(align_bam_file,'rb')

    info_dict=extract_pairs_pos(bam_file,position,length,chrom,strand)
    if info_dict == {}:
        raise RuntimeError("There is no read aligned on this position")
    info_df = pd.DataFrame(list(info_dict.keys()))

    s5 = pyslow5.Open(slow5_file, 'r')
    read_ids, num_reads = s5.get_read_ids()
    df=pd.read_csv(paf_file,sep='\t',header=None)
    df=pd.merge(df,info_df,how='inner',on=0)
    if df.shape[0] == 0:
        raise RuntimeError("cannot found the record from bam in your paf file. Please check your f5c command ... ")
    if df.shape[0] / info_df.shape[0] < 0.8:
        print('There are '+str(info_df.shape[0]-df.shape[0])+" reads not found in your paf file ...")
    pbar = tqdm(total=df.shape[0], position=0, leave=True)
    df["feature"] = df.apply(extract_feature,strand=strand,norm=norm,sig_move_offset=sig_move_offset,axis=1)
    pbar.close()

    df.dropna(inplace=True)
    num_aligned = df.shape[0]
    if subsample_ratio<1:
        df=df.sample(frac=subsample_ratio)
    final_feature=[]
    for item in df["feature"]:
        final_feature.extend(item)
    final_feature=pd.DataFrame(final_feature)
    final_feature.columns=['Read ID','Mean','STD','Median','Dwell time','Position']
    # if rna_mode:
    #     if strand == '+':
    #         final_feature['position']=final_feature['position'] - (kmer_model-1)
    #     else:
    #         final_feature['position']=final_feature['position'] + (kmer_model-1)

    final_feature['Position'] = final_feature['Position'].astype(int).astype(str)
    print('Extracted ', num_aligned, ' aligned reads from blow5 files')

    # if num_aligned>50:
    #     dwell_filter_pctls = (0.5, 99.5)
    #     dwell_min, dwell_max = np.percentile(final_feature['Dwell time'].values, dwell_filter_pctls)
    #     final_feature = final_feature[(final_feature['Dwell time'] > dwell_min) & (final_feature['Dwell time'] < dwell_max)]

    return final_feature,num_aligned





import numpy as np
from matplotlib import pyplot as plt
import pandas as pd
import xgboost
from nanoCEM.cem_utils import  extract_kmer_feature
plt.rcParams['pdf.fonttype'] = 42
results_path='f5c_result_rna'
df = pd.read_csv(results_path+'/current_feature.csv')

feature_matrix, label = extract_kmer_feature( df, 7, 2030)

label[0] = label[0].apply(lambda x: 1 if x == 'Sample' else 0)

X = feature_matrix.values
y = label.values

from sklearn.metrics import accuracy_score
from sklearn.model_selection import KFold
kfold = KFold(n_splits=5)
# Perform k-fold cross-validation
accuracy_scores=[]
for train_index, test_index in kfold.split(X):
    # Split the data into training and testing sets for each fold
    X_train, X_test = X[train_index], X[test_index]
    y_train, y_test = y[train_index], y[test_index]

    # Create an XGBoost classifier
    model = xgboost.XGBClassifier()

    # Train the model on the training data
    model.fit(X_train, y_train)

    # Make predictions on the testing data
    y_pred = model.predict(X_test)

    # Calculate the accuracy score
    accuracy = accuracy_score(y_test, y_pred)

    # Store the accuracy score for this fold
    accuracy_scores.append(accuracy)

    # Calculate the average accuracy across all folds
avg_accuracy = sum(accuracy_scores) / len(accuracy_scores)
print(avg_accuracy)



y_pred = model.predict(X_test)
accuracy = accuracy_score(y_test, y_pred)
# explain the model's predictions using SHAP
# (same syntax works for LightGBM, CatBoost, scikit-learn, transformers, Spark, etc.)
print(accuracy)
y_pred = model.predict_proba(X_test)[:,1]
prediction_df = pd.DataFrame({'y_test': y_test.reshape(-1,), 'y_pred': y_pred})
prediction_df['y_test'] = prediction_df['y_test'].apply(lambda x: 'Sample' if x==1 else 'Control')
import plotnine as p9
category = pd.api.types.CategoricalDtype(categories=['Sample', "Control"], ordered=True)
prediction_df['y_test'] = prediction_df['y_test'].astype(category)
# visualize the first prediction's explanation
plot = p9.ggplot(prediction_df, p9.aes(x='y_test', y="y_pred",fill='y_test')) \
            +p9.scale_fill_manual(values={"Sample": "#F57070", "Control": "#9F9F9F", "Single": "#a3abbd"})\
           + p9.theme_bw() \
           + p9.labs(x='',y='Prediction')\
           + p9.geom_boxplot(width=0.6,alpha=0.7)\
           + p9.theme(
        figure_size=(4, 4),
        panel_grid_minor=p9.element_blank(),
        axis_text=p9.element_text(size=13),
        axis_title=p9.element_text(size=13),
        title=p9.element_text(size=13),
        legend_position='none',
        legend_title=p9.element_blank(),
        strip_text=p9.element_text(size=13),
        strip_background=p9.element_rect(alpha=0),
    )
print(plot)
plot.save(filename=results_path + "/prediction_barplot.pdf", dpi=300)
plot = p9.ggplot(prediction_df, p9.aes(fill='y_test', x="y_pred")) \
    +p9.scale_fill_manual(values={"Sample": "#F57070", "Control": "#9F9F9F", "Single": "#a3abbd"})\
           + p9.theme_bw() \
           + p9.labs(x='Prediction',y='Density')\
           + p9.geom_density(alpha=0.7)\
           + p9.theme(
        figure_size=(4,4),
        panel_grid_minor=p9.element_blank(),
        axis_text=p9.element_text(size=13),
        axis_title=p9.element_text(size=13),
        title=p9.element_text(size=13),
        legend_position='bottom',
        legend_title=p9.element_blank(),
        strip_text=p9.element_text(size=13),
        strip_background=p9.element_rect(alpha=0),
    )
print(plot)

plot.save(filename=results_path + "/prediction_distribution.pdf", dpi=300)



# 示例数据




import re
import numpy as np
import pandas as pd
import pysam
import pyslow5
from tqdm import tqdm

from nanoCEM.cem_utils import generate_bam_file, identify_file_path, generate_paf_file_resquiggle, base_shift_dict,ker_model_size
from nanoCEM.normalization import normalize_signal_with_lim
# from nanoCEM.plot import draw_signal
from nanoCEM.read_move_table import extract_pairs_pos

# import os
# import argparse
score_dict = {}
nucleotide_type = None

def extract_feature(line,kmer_model,strand,base_shift, norm=True):
    global nucleotide_type
    pbar.update(1)
    read_id = line[0]
    # if read_id =='94c4929a-8630-4fd0-89d5-21f6bc399165':
    #     print(1)
    if read_id not in info_dict:
        return None
    # tackle moves tag
    moves_string = line[14]
    moves_string = re.sub('ss:Z:', '', moves_string)
    moves_string = re.sub('D', 'D,', moves_string)
    moves_string = re.sub('I', 'I,', moves_string)
    # print(moves_string)
    moves = re.split(r',+', moves_string)
    moves = moves[:-1]
    # extract index and generate event_length and event_start
    insertion = 0
    event_length = []
    for i, item in enumerate(moves):
        if 'D' in item:
            deletion = int(item[:-1])
            for i in range(deletion):
                event_length.append(0)
        elif 'I' in item:
            if i == 0:
                continue
            else:
                return None
        elif '=' in item:
            return None
        else:
            event_length.append(int(item))
    # build event_length from move table
    read = s5.get_read(read_id, aux=["read_number", "start_mux"], pA=True)
    start_index = line[2]
    end_index = line[3]
    event_length = np.array(event_length)

    # identify RNA or DNA
    if nucleotide_type is None:
        if line[7] > line[8]:
            nucleotide_type = 'RNA'
        else:
            nucleotide_type = 'DNA'
    #  assert len_raw_signal in paf and blow5
    try:
        assert end_index - start_index == np.sum(event_length)
        assert read['len_raw_signal'] == line[1]

    except Exception:
        print("Warning: 1 read's length of signal is not equal between blow5 and paf")
        return None

    signal = read['signal']

    # create event start and flip all table
    signal = signal[start_index:end_index]
    event_starts = event_length.cumsum()
    event_starts = np.insert(event_starts, 0, 0)[:-1]
    if norm:
        signal, _, _ = normalize_signal_with_lim(signal)

    # index query and reference
    aligned_pair = info_dict[read_id]['pairs']
    qlen = info_dict[read_id]['query_length']
    # assert len(event_length) + kmer_model - 1 == qlen

    # correct index about DNA and RNA
    if (nucleotide_type == 'RNA' and strand=='+') or (nucleotide_type == 'DNA' and strand=='-'):
        aligned_pair[0] = qlen - aligned_pair[0] - 1
        aligned_pair[1] = aligned_pair[1] - kmer_model - base_shift + 1
    else:
        aligned_pair[1] = aligned_pair[1] - base_shift

    aligned_pair = aligned_pair[(aligned_pair[0] >= 0) & (aligned_pair[0] < len(event_length))]
    if aligned_pair.shape[0] == 0:
        return None
    read_pos = aligned_pair[0].values
    ref_pos = aligned_pair[1].values

    # extract raw signal by event length and event start
    total_feature_per_reads = []
    try:
        raw_signal_every = [signal[event_starts[x]:event_starts[x] + event_length[x]] for x in
                            read_pos]
        # if aligned_pair.shape[0] < 11:
        #     draw_signal(signal[event_starts[read_pos[0]]:event_starts[read_pos[-1]+1]], event_starts[read_pos[0]:read_pos[-1]+1]-event_starts[read_pos[0]], base_list)
    except Exception as e:
        print(e)
        return None
    # calculate mean median and dwell time
    for i, element in enumerate(raw_signal_every):
        if event_length[read_pos[i]] == 0:
            continue
        temp = [read_id, np.mean(element), np.std(element), np.median(element), event_length[read_pos[i]], ref_pos[i]]
        total_feature_per_reads.append(temp)
    return total_feature_per_reads


def read_blow5(path, position, reference, length, chrom, strand, pore, subsample_ratio=1, base_shift=True, norm=True,
               cpu=4, rna=True):
    global s5, pbar, info_dict
    slow5_file = path + ".blow5"
    fastq_file = path + ".fastq"
    identify_file_path(fastq_file)
    identify_file_path(slow5_file)

    if rna:
        nucleotide_type = 'RNA'
    else:
        nucleotide_type = 'DNA'
    kmer_model = ker_model_size[pore+nucleotide_type]
    if base_shift == 'auto':
        base_shift = base_shift_dict[pore + nucleotide_type + strand]
    else:
        base_shift = base_shift

    fastq_file, bam_file = generate_bam_file(fastq_file, reference, cpu, subsample_ratio)
    paf_file = generate_paf_file_resquiggle(fastq_file, slow5_file, pore, rna, cpu)
    bam_file = pysam.AlignmentFile(bam_file, 'rb')
    info_dict = extract_pairs_pos(bam_file, position, length + abs(kmer_model), chrom, strand)
    if info_dict == {}:
        raise RuntimeError("There is no read aligned on this position")
    info_df = pd.DataFrame(list(info_dict.keys()))

    s5 = pyslow5.Open(slow5_file, 'r')
    df = pd.read_csv(paf_file, sep='\t', header=None)
    df = pd.merge(df, info_df, how='inner', on=0)
    if df.shape[0] == 0:
        raise RuntimeError("cannot found the record from bam in your paf file. Please check your f5c command ... ")
    if df.shape[0] / info_df.shape[0] < 0.8:
        print('There are ' + str(info_df.shape[0] - df.shape[0]) + " reads not found in your paf file ...")
    pbar = tqdm(total=df.shape[0], position=0, leave=True)
    df["feature"] = df.apply(extract_feature,kmer_model=kmer_model, strand = strand,base_shift=base_shift, norm=norm, axis=1)
    pbar.close()

    df.dropna(inplace=True)
    num_aligned = df.shape[0]
    final_feature = []
    for item in df["feature"]:
        final_feature.extend(item)
    final_feature = pd.DataFrame(final_feature)
    final_feature.columns = ['Read ID', 'Mean', 'STD', 'Median', 'Dwell time', 'Position']
    final_feature = final_feature[
        (final_feature['Position'] >= position - length) & (final_feature['Position'] <= position + length)]

    final_feature['Position'] = final_feature['Position'].astype(int).astype(str)
    print('Extracted ', num_aligned, ' aligned reads from blow5 files')

    # if num_aligned>50:
    #     dwell_filter_pctls = (0.5, 99.5)
    #     dwell_min, dwell_max = np.percentile(final_feature['Dwell time'].values, dwell_filter_pctls)
    #     final_feature = final_feature[(final_feature['Dwell time'] > dwell_min) & (final_feature['Dwell time'] < dwell_max)]

    return final_feature, num_aligned, nucleotide_type



#!/usr/bin/env python

# !/usr/bin/env python
import argparse
import copy
import os
import time
import warnings

import pandas as pd
from plotnine.exceptions import PlotnineWarning

from nanoCEM.cem_utils import read_fasta_to_dic,identify_file_path,build_out_path
from nanoCEM.plot import current_plot,plot_PCA

warnings.filterwarnings("ignore", category=PlotnineWarning)
from nanoCEM.cem_utils import generate_bam_file

def init_parser():
    def add_public_argument(parser_input):
        parser_input.add_argument("--chrom", required=True, help="Gene or chromosome name(head of your fasta file)")
        parser_input.add_argument("--pos", required=True, type=int, help="site of your interest")
        parser_input.add_argument("--len", default=10, type=int, help="region around the position")
        parser_input.add_argument("--strand", default="+", help="Strand of your interest")
        parser_input.add_argument('-r', "--ref", required=True, help="fasta file")
        parser_input.add_argument('-t', "--cpu", default=4, type=int, help="num of process")
        parser_input.add_argument('--norm', action='store_true', help='Turn on the normalization mode')
        parser_input.add_argument('-s', "--subsample_ratio", default=1, type=float,
                                  help="Subsample ratio to select reads")
        parser_input.add_argument('--rna', action='store_true', help='Turn on the RNA mode')
        parser_input.add_argument('-o', "--output", default="nanoCEM_result", help="output_file")

    # Define the argument parser
    parser = argparse.ArgumentParser(
        description='A sample tool designed to visualize the features that distinguish between two groups of ONT data at the site level. It supports two re-squiggle pipeline(Tombo and f5c).')
    subparsers = parser.add_subparsers(dest='function')

    # tombo subparser
    parser_tombo = subparsers.add_parser('tombo', help='tackle tombo re-squiggle')
    parser_tombo.add_argument('--basecall_group', default="RawGenomeCorrected_000",
                              help='The attribute group to extract the training data from. e.g. RawGenomeCorrected_000')
    parser_tombo.add_argument('--basecall_subgroup', default='BaseCalled_template',
                              help='Basecall subgroup Nanoraw resquiggle into. Default is BaseCalled_template')
    parser_tombo.add_argument('-i', "--input_fast5", required=True,
                              help="input_fast5_file")
    parser_tombo.add_argument('-c', "--control_fast5",
                              help="control_fast5_file")
    # parser_tombo.add_argument('-b', "--bam", help="bam file to help index to speed up")


    add_public_argument(parser_tombo)

    # f5c subparser
    parser_f5c = subparsers.add_parser('f5c', help='tackle f5c re-squiggle')
    parser_f5c.add_argument("-i", "--input", required=True,
                            help="blow5_path")
    parser_f5c.add_argument('-c', "--control",
                            help="control_blow5_path")
    parser_f5c.add_argument('--base_shift', type=int, default=2, help="used for result shifting in f5c")
    add_public_argument(parser_f5c)
    # parser.set_defaults(function='tombo', chrom="NR_103073.1", pos=2030, len=10, strand='+', cpu=8,norm=True,
    #                     input_fast5='../example/data/wt/single/', \
    #                     control_fast5='../example/data/ivt/single/', output='tombo_result_rna',
    #                     subsample_ratio=1, \
    #                     ref="../example/data/23S_rRNA.fasta", \
    #                     basecall_group="RawGenomeCorrected_000", basecall_subgroup="BaseCalled_template", rna=True)
    # parser.set_defaults(function='tombo', chrom="NR_103073.1", pos=875, len=10, strand='-', cpu=8,norm=True,
    #                     input_fast5='../example/data/reverse/wt/single/', \
    #                     control_fast5='../example/data/reverse/ivt/single/', output='tombo_result_rna_re',
    #                     subsample_ratio=1, \
    #                     ref="../example/data/reverse/23S_rRNA_re.fasta", \
    #                     basecall_group="RawGenomeCorrected_000", basecall_subgroup="BaseCalled_template", rna=True)
    # parser.set_defaults(function='f5c', chrom="NR_103073.1", pos=2030, len=10, strand='+', cpu=8,norm=True,base_shift=2,
    #                     input='../example/data/wt/file', \
    #                     control='../example/data/ivt/file', output='f5c_result_rna',
    #                     subsample_ratio=1, \
    #                     ref="../example/data/23S_rRNA.fasta", \
    #                     basecall_group="RawGenomeCorrected_000", basecall_subgroup="BaseCalled_template", rna=True)
    parser.set_defaults(function='f5c', chrom="NR_103073.1", pos=875, len=10, strand='-', cpu=8,norm=True,base_shift=2,
                        input='../example/data/reverse/wt/wt', control='../example/data/reverse/ivt/ivt', output='f5c_result_rna_re',
                        subsample_ratio=1,
                        ref="../example/data/reverse/23S_rRNA_re.fasta", rna=True)
    return parser

if __name__ == '__main__':
    # Parse the arguments
    parser = init_parser()
    args = parser.parse_args()

    # args check
    if args.function is None:
        raise Exception("Please choose function from f5c or tombo, or use -h to view the help document")
    identify_file_path(args.ref)
    fasta = read_fasta_to_dic(args.ref)
    args.pos = args.pos - 1
    length_gene = len(fasta[args.chrom])
    if args.pos + args.len + 1 >= length_gene or args.pos - args.len <= 0:
        raise Exception("The position requested is too close to the border (pos-len>0 and pos+len<length of fasta)")

    # build title
    base_list = fasta[args.chrom][args.pos - args.len:args.pos + args.len + 1].upper()
    title = args.chrom + ':' + str(args.pos - args.len + 1) + '-' + str(args.pos + args.len + 1) + ':' + args.strand
    if args.rna:
        base_list = base_list.replace("T", "U")
        print("Running RNA mode ...")
        title = 'RNA  ' + title
    else:
        print("Running DNA mode ...")
        title = 'DNA  ' + title

    results_path = args.output
    build_out_path(results_path)
    subsample_ratio = args.subsample_ratio
    if args.function == 'tombo':
        from nanoCEM.read_tombo_resquiggle import create_read_list_file, extract_group

        wt_file = create_read_list_file(args.input_fast5, results_path)
        df_wt, aligned_num_wt = extract_group(args, wt_file, subsample_ratio)
        df_wt['Group'] = 'Sample'
        try:
            ivt_file = create_read_list_file(args.control_fast5, results_path)
            df_ivt, aligned_num_ivt = extract_group(args, ivt_file, subsample_ratio)
            df_ivt['Group'] = 'Control'
            df = pd.concat([df_wt, df_ivt])
            title = title + '   Sample:' + str(aligned_num_wt) + '  Control:' + str(aligned_num_ivt)
            category = pd.api.types.CategoricalDtype(categories=['Sample', "Control"], ordered=True)
            df['Group'] = df['Group'].astype(category)
        except:
            args.control_fast5 = None
        if args.control_fast5 is None:
            df = df_wt
            df_wt['Group'] = 'Single'
            title = title + '   Sample:' + str(aligned_num_wt)
    elif args.function == 'f5c':
        from nanoCEM.read_f5c_resquiggle import read_blow5

        if args.base_shift < 0:
            raise RuntimeError("base_shift should not less than 0")
        # if (args.rna and args.strand=='+') or (not args.rna and args.strand=='-'):
        #     args.pos = args.pos + args.base_shift
        # else:
        #     args.pos = args.pos - args.base_shift

        df_wt, aligned_num_wt, nucleotide_type = read_blow5(args.input, args.pos,args.ref, args.len, args.chrom, args.strand,
                                                            subsample_ratio, args.base_shift, args.norm, args.cpu)
        df_wt['Group'] = 'Sample'
        if nucleotide_type == 'RNA' and not args.rna:
            raise RuntimeError("You need to add --rna to turn on the rna mode")
        try:
            df_ivt, aligned_num_ivt, _ = read_blow5(args.control, args.pos,args.ref, args.len, args.chrom, args.strand,
                                                    subsample_ratio, args.base_shift, args.norm, args.cpu)
            df_ivt['Group'] = 'Control'

            df = pd.concat([df_wt, df_ivt])
            category = pd.api.types.CategoricalDtype(categories=['Sample', "Control"], ordered=True)
            df['Group'] = df['Group'].astype(category)

            title = title + '   Sample:' + str(aligned_num_wt) + '  Control:' + str(aligned_num_ivt)
        except:
            args.control = None
        if args.control is None:
            df = df_wt
            df_wt['Group'] = 'Single'
            title = title + '   Sample:' + str(aligned_num_wt)

    # draw_volin(df,results_path,args.pos,base_list,title)
    # draw_boxplot(df,results_path,args.pos,base_list,title)

    df_copy = copy.deepcopy(df)
    df_copy['Position'] = df_copy['Position'].astype(int) + 1
    df_copy.to_csv(results_path + '/current_feature.csv', index=None)
    print("Feature file saved  in " + results_path + '/current_feature.csv')

    # percentile_filter = False
    # if aligned_num_wt > 50 and aligned_num_ivt > 50:
    #     percentile_filter = True

    current_plot(df, results_path, args.pos, base_list, title)
    # signal_plot(df, results_path, args.pos, base_list, title, 'boxplot')
    # signal_plot(df, results_path, args.pos, base_list, title, 'violin_plot')
    print('Finished')

    df = df_copy[df_copy['Position'] == args.pos + 1]
    plot_PCA(df, results_path)



import numpy as np
OUTLIER_THRESH = 5.0
MAX_POINTS_FOR_THEIL_SEN = 1000

def normalize_signal_with_lim(raw_signal, lower_lim=-OUTLIER_THRESH, upper_lim=OUTLIER_THRESH):
    # shift = np.median(signal)
    # # note factor to scale MAD to SD
    # scale = np.median(np.absolute(signal - shift)) * 1.4826
    # norm_signal = (signal - shift) / scale
    shift = np.median(raw_signal)
    scale = np.median(np.abs(raw_signal - shift))
    norm_signal = (raw_signal - shift) / scale

    # norm_signal = c_apply_outlier_thresh(
    #         norm_signal, lower_lim, upper_lim)

    norm_signal = np.where(norm_signal > upper_lim, upper_lim, norm_signal)
    norm_signal = np.where(norm_signal < lower_lim, lower_lim, norm_signal)
    return norm_signal,shift,scale

def normalize_signal(raw_signal):
    # shift = np.median(signal)
    # # note factor to scale MAD to SD
    # scale = np.median(np.absolute(signal - shift)) * 1.4826
    shift = np.median(raw_signal)
    scale = np.median(np.abs(raw_signal - shift))
    norm_signal = (raw_signal - shift) / scale

    # norm_signal = c_apply_outlier_thresh(
    #         norm_signal, lower_lim, upper_lim)
    return norm_signal


def calc_kmer_fitted_shift_scale(prev_shift, prev_scale, r_event_means, r_model_means,):
    """Use robust Theil-Sen estimator to compute fitted shift and scale
    parameters based on read sequence

    Args:
        prev_shift (float): previous shift parameter
        prev_scale (float): previous scale parameter
        r_ref_means (`np.array::np.float64`): expected base signal levels
        r_ref_sds (`np.array::np.float64`): expected base signal level sds


    Returns:
        Sequence-fitted scaling parameters

        1) shift parameter (float)
        2) scale parameter (float)
        3) shift correction factor; multiply by ``prev_scale`` and add to ``prev_shift`` to get ``shift`` (float)
        4) scale correction factor; multiply by ``prev_scale`` to get ``scale`` (float)
    """
    def compute_slopes(r_event_means, r_model_means):
        # despite computing each diff twice this vectorized solution is
        # about 10X faster than a list comprehension approach
        delta_event = r_event_means[:, np.newaxis] - r_event_means
        delta_model = r_model_means[:, np.newaxis] - r_model_means
        return delta_model[delta_event > 0] / delta_event[delta_event > 0]

    n_points = r_model_means.shape[0]
    # potentially sample points for long reads (>1kb)
    if r_model_means.shape[0] > MAX_POINTS_FOR_THEIL_SEN:
        n_points = MAX_POINTS_FOR_THEIL_SEN
        samp_ind = np.random.choice(
            r_model_means.shape[0], n_points, replace=False)
        r_model_means = r_model_means[samp_ind]
        r_event_means = r_event_means[samp_ind]
    # compute Theil-Sen slope estimator
    slope = np.median(compute_slopes(r_event_means, r_model_means))
    inter = np.median(r_model_means - (slope * r_event_means))
    if slope == 0:
        raise RuntimeError('Read failed sequence-based signal ' +
                            're-scaling parameter estimation.')
        # convert to shift and scale parameters (e.g. (obs - shift) / scale)
        scale_corr_factor = 1 / slope
        shift_corr_factor = -inter / slope

    # apply shift and scale values fitted from kmer conditional model
    shift = prev_shift + (shift_corr_factor * prev_scale)
    scale = prev_scale * scale_corr_factor
    return shift, scale, shift_corr_factor, scale_corr_factor

# from nanoCEM._c_helper import c_new_means
# def compute_base_means(all_raw_signal, base_starts):
#     """Efficiently compute new base mean values from raw signal and base start
#     positions
#
#     Args:
#         all_raw_signal (`np.array`): raw nanopore signal obervation values
#         base_starts (`np.array::np.int32`): 0-based base start positions within
#             raw signal
#
#     Returns:
#         `np.array::np.float64` containing base mean levels
#     """
#     return c_new_means(all_raw_signal.astype(np.float64), base_starts)
#
# def normal_new(signal,segs):
#     norm_signal,shift,scale = normalize_signal_with_lim
#     (shift, scale, shift_corr_factor,
#      scale_corr_factor) = calc_kmer_fitted_shift_scale(
#         shift, scale,
#         compute_base_means(norm_signal, segs))
#     # re-normalize signal with new fitted parameters
#     norm_signal = (norm_signal - shift_corr_factor) / scale_corr_factor
